;---------------------------------------------------------------------
; Criba de Eratóstenes
; Calcula los números primos entre 2 y MAXPR
;---------------------------------------------------------------------
;---------------------------------------------------------------------
; Convenciones usadas en este programa:
; - R7 es el puntero de pila.
; - R6 se usa para guardar la dirección de retorno en las subrutinas
; - La pila es del tipo "Full-Descending"
; - R0, R1 y R2 se usan para pasar parámetros y retornar resultados.
;               No se necesitan preservar en las subrutinas
; - R3, R4 y R5 se usan para las variables locales.
;               Su valor se debe preservar en las subrutinas
;---------------------------------------------------------------------
UARTDAT=	0x22
PFLAGS=		0x24

MAXPR=		32768

;--------- Vectores ---------
		org 0
		jr		inicio

;--------- CODE -----------
		org		0x40
inicio:
		ldpc	r7				; puntero de pila inicial
		word	4096

		ldi		r0,0xff			; Sync UART PC
		adpc	r6,1
		jr		putch
		; rellenamos inicialmente la criba con ceros
		; el buffer está en la memoria de datos a partir de TABBASE
		ldpc	r1
		word	(MAXPR/16)
		ldpc	r2
		word	TABBASE
		ldi		r0,0
l1:		st		(r2),r0
		addi	r2,r2,1
		subi	r1,r1,1
		jnz		l1

		ldi		r5,10	;// nº de caracteres por línea
		; R3: número bajo prueba
		ldi		r3,2
bmain:	; probamos el número
		shr		r1,r3		; R1=R3>>4
		shr		r1,r1
		shr		r1,r1
		shr		r1,r1
		ldpc	r2
		word	TABBASE
		add		r1,r2,r1
		ld		r0,(r1)
		andi	r2,r3,0xF	; R2=bit a probar
		jz		l3
l2:		shr		r0,r0
		subi	r2,r2,1
		jnz		l2
l3:		tsti	r0,1		; si 0 el número es primo
		jnz		noprim

		; Imprimimos el número
		ldi		r4,0		; r4=contador de dígitos	
		or		r0,r3,r3	; r0=dividendo
prnum:	ldi		r1,0		; r1=resto
		ldi		r2,16		; r2=contador de bits
l4:		add		r0,r0,r0
		adc		r1,r1,r1
		cmpi	r1,10
		jnc		l5
		subi	r1,r1,10
		ori		r0,r0,1
l5:		subi	r2,r2,1
		jnz		l4

		subi	r7,r7,1		; resto a la pila
		st		(r7),r1
		addi	r4,r4,1
		cmpi	r0,0
		jnz		prnum		; siguiente dígito

l6:		ld		r0,(r7)		; dato desde la pila
		addi	r7,r7,1
		ldi		r1,'0'
		add		r0,r0,r1	; a ASCII
		adpc	r6,1
		jr		putch
		subi	r4,r4,1
		jnz		l6

		; imprimimos espacio o NL
		ldi		r0,' '
		subi	r5,r5,1
		jnz		l65
		ldi		r5,10
		ldi		r0,'\n'
l65:	adpc	r6,1
		jr		putch
		; y marcamos sus múltiplos como no primos, empezando por el 2º
		add		r4,r3,r3	; r4=múltiplo
l7:		jc		noprim		; acarreo => mayor de 65536
		ldpc	r0
		word	MAXPR
		cmp		r4,r0		; si r4>MAXPR terminamos con los múltiplos
		jpl		noprim
		; Marcamos bit en la criba (bit en 1 => no primo)
		shr		r1,r4		; R1=R4>>4
		shr		r1,r1
		shr		r1,r1
		shr		r1,r1
		ldpc	r2
		word	TABBASE
		add		r1,r2,r1
		ldi		r0,1		; R0= máscara
		andi	r2,r4,0xF	; R2= bit a borrar
		jz		l9
l8:		add		r0,r0,r0	; R0<<=1
		subi	r2,r2,1
		jnz		l8
l9:		ld		r2,(r1)
		or		r2,r2,r0
		st		(r1),r2
		add		r4,r4,r3	; siguiente múltiplo
		jr		l7
noprim:	addi	r3,r3,1		; siguiente número
		ldpc	r0
		word	MAXPR
		cmp		r3,r0
		jnz		bmain

		ldi		r0,'\n'
		adpc	r6,1
		jr		putch

		;word	0x5880		; ILEG
		jr		.


;----------------------------------------------
; Envía caracter al terminal
; parámetros:
;	R0: dato a enviar
;	r6: dirección de retorno
; retorna:
; 	R1, R2: modificados

putch:
		;jind	r6		; dummy putch

		ldi		r1,PFLAGS
		; esperamos flag tx en 1 (listo para TX)
putch1:	ld		r2,(r1)
		tsti	r2,2
		jz		putch1
		ldi		r1,UARTDAT	; ahora apuntamos a registro TX
		st		(r1),r0		; envía dato
		jind	r6			; y retornamos

TABBASE:	; memoria libre para la tabla


