;---------------------------------------------------------------------
; Convenciones usadas en este programa:
; - R7 es el puntero de pila.
; - R6 se usa para guardar la dirección de retorno en las subrutinas
; - La pila es del tipo "Full-Descending"
; - R0, R1 y R2 se usan para pasar parámetros y retornar resultados.
;               No se necesitan preservar en las subrutinas
; - R3, R4 y R5 se usan para las variables locales.
;               Su valor se debe preservar en las subrutinas
;---------------------------------------------------------------------
;----------- REGISTROS de E/S -----------

IRQEN=		0x20
TIMER=		0x21
UARTDAT=	0x22
SPIDAT=		0x23
PFLAGS=		0x24
SPICTRL= 	0x25
GPIN=		0x25
GPDIR=		0x26
GPOUT=		0x27

;--------- Habilitación de interrupciones --------
EINTEX=		1		; Habilita IRQ externa
ETIMER=		2		; Habilita IRQ Timer
EUARTTHRE=	4		; Habilita IRQ UART TX
EUARTDV=	8		; Habilita IRQ UART RX

;--------- bits de PFLAGS -----------
UARTDV=		1		; Dato válido en UART RX
UARTTHRE=	2		; Listo para transmitir en UART TX
UARTTEND=	4		; Transmisión completa en UART TX
SPIBUSY=	8		; SPI ocupado transfiriendo datos
UARTFE=		0x10	; Error de trama (bit de STOP en 0) en UART RX
UARTOVE=	0x20	; Error de overrun en UART RX
TIMERIRQ=	0x8000	; Flag de IRQ del temporizador

;------------- Notas E/S --------------------
; - La escritura en TIMER carga el valor máximo de la cuenta y pone el contador en 0
; - La lectura de TIMER devuelve el valor actual de la cuenta y borra la petición de interrupción
; - El bit 8 de SPICTRL selecciona el modo de 16 bits si 1 o 8 bits si 0.
; - Los bits 7:0 de SPICTRL seleccionan la velocidad de SCK como f_sck=f_clk/(2*(SPICTRL[7:0]+1))

		org 0			; RESET
		jr	inicio

		org 4			; IRQ0: externa
		JR		IRQ0
		org 8			; IRQ1: TIMER
		jr		IRQ1
		org	0xc			; IRQ2:	UART TX
		jr		IRQ2
		org	0x10		; IRQ3: UART RX
		jr		IRQ3

		org	64
IRQ1:	subi	r7,r7,1
		st		(r7),r0
		subi	r7,r7,1
		st		(r7),r1

		ldi		r1,TIMER	; La lectura borra la IRQ del timer
		ld		r0,(r1)

		ldpc	r1
		word	outreg
		ld		r0,(r1)
		addi	r0,r0,1
		st		(r1),r0
		ldi		r1,<GPOUT
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		st		(r1),r0

		ld		r1,(r7)
		addi	r7,r7,1
		ld		r0,(r7)
		addi	r7,r7,1
IRQ0:
IRQ2:
IRQ3:
		reti
;-------------------------------------------------------------------
;
;		INICIO
;
;-------------------------------------------------------------------
inicio:

		ldpc	r7		; Puntero de Pila al final de RAM
		word	4096

		; GPIO como salida
		ldi		r1,GPDIR
		ldi		r0,0x1F
		st		(r1),r0 

		; Prueba de interrupciones 

		ldi		r1,TIMER	; Interrupción cada 100 ciclos
		ldpc	r0
		word	9999
		st		(r1),r0
		ld		r0,(r1)		; Borramos IRQ
		ldi		r1,IRQEN	; Habilita IRQ Timer
		ldi		r0,ETIMER
		st		(r1),r0

		ldpc	r0
		word	txtle
		adpc	r6,1
		jr		putsle

		ldpc	r4
		word	0x7FE0
		ldi		r5,0x0F
buc:
		or		r0,r5,r5
		adpc	r6,1
		jr		prthex
		ldi		r0,'+'
		adpc	r6,1
		jr		putch

		or		r0,r4,r4
		adpc	r6,1
		jr		prthex
		ldi		r0,'='
		adpc	r6,1
		jr		putch

		add		r0,r5,r4
		adpc	r6,1
		jr		prthex
		ldi		r0,' '
		adpc	r6,1
		jr		putch

		add		r0,r5,r4
		jv		.+2
		jr		novf
		ldi		r0,'V'
		adpc	r6,1
		jr		putch

novf:	ldi		r0,'\n'
		adpc	r6,1
		jr		putch

		addi	r4,r4,1
		jpl		buc


		ldpc	r4
		word	0xFFE0
		ldpc	r5
		word	0x8010
buc1:
		or		r0,r5,r5
		adpc	r6,1
		jr		prthex
		ldi		r0,'+'
		adpc	r6,1
		jr		putch

		or		r0,r4,r4
		adpc	r6,1
		jr		prthex
		ldi		r0,'='
		adpc	r6,1
		jr		putch

		add		r0,r5,r4
		adpc	r6,1
		jr		prthex
		ldi		r0,' '
		adpc	r6,1
		jr		putch

		add		r0,r5,r4
		jv		.+2
		jr		novf1
		ldi		r0,'V'
		adpc	r6,1
		jr		putch

novf1:	ldi		r0,'\n'
		adpc	r6,1
		jr		putch

		addi	r4,r4,1
		jmi		buc1


		ldi		r4,0
		ldpc	r5
		word	0x8008

buc2:
		or		r0,r5,r5
		adpc	r6,1
		jr		prthex
		ldi		r0,'-'
		adpc	r6,1
		jr		putch

		or		r0,r4,r4
		adpc	r6,1
		jr		prthex
		ldi		r0,'='
		adpc	r6,1
		jr		putch

		sub		r0,r5,r4
		adpc	r6,1
		jr		prthex
		ldi		r0,' '
		adpc	r6,1
		jr		putch

		sub		r0,r5,r4
		jv		.+2
		jr		novf2
		ldi		r0,'V'
		adpc	r6,1
		jr		putch

novf2:	ldi		r0,'\n'
		adpc	r6,1
		jr		putch

		addi	r4,r4,1
		cmpi	r4,15
		jnz		buc2


		ldpc	r4
		word	0xFFE0
		ldpc	r5
		word	0x7FF0

buc3:
		or		r0,r5,r5
		adpc	r6,1
		jr		prthex
		ldi		r0,'-'
		adpc	r6,1
		jr		putch

		or		r0,r4,r4
		adpc	r6,1
		jr		prthex
		ldi		r0,'='
		adpc	r6,1
		jr		putch

		sub		r0,r5,r4
		adpc	r6,1
		jr		prthex
		ldi		r0,' '
		adpc	r6,1
		jr		putch

		sub		r0,r5,r4
		jv		.+2
		jr		novf3
		ldi		r0,'V'
		adpc	r6,1
		jr		putch

novf3:	ldi		r0,'\n'
		adpc	r6,1
		jr		putch

		addi	r4,r4,1
		jmi		buc3

		jr	.


;----------------------------------------------------
; Envía cadena de caracteres empaquetados al terminal
; parámetros:
;	R0: puntero a la cadena de caracteres
;	R6: dirección de retorno
; retorna:
; 	R0-R2: modificados

putsle:	subi	r7,r7,1
		st		(r7),r6

		or		r2,r0,r0
putsle1:
		ld		r0,(r2)
		ldi		r1,0xff
		and		r0,r0,r1
		jz		putsle3
		adpc	r6,1
		jr		putch

		ld		r0,(r2)
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		jz		putsle3
		adpc	r6,1
		jr		putch
		addi	r2,r2,1
		jr		putsle1

putsle3:
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6
;----------------------------------------------
; Envía caracter al terminal
; parámetros:
;	R0: dato a enviar
;	R6: dirección de retorno
; retorna:
; 	R1: modificado

putch:	ldi		r1,PFLAGS	; Esperamos flag TBE en 1
		ld		r1,(r1)		
		tsti	r1,UARTTHRE
		jz		putch
		ldi		r1,<UARTDAT		
		st		(r1),r0		; envía dato
		jind	r6			; y retornamos


;----------------------------------------------
; Recibe caracter del terminal
; parámetros:
;	R6: dirección de retorno
; retorna:
;	R0: dato recibido
; 	R1: modificado

getch:	ldi		r1,PFLAGS
getch1:	ld		r0,(r1)		; esperamos flag DV en 1
		tsti	r0,UARTDV
		jz		getch1
		ldi		r1,UARTDAT	; ahora apuntamos a registro RX
		ld		r0,(r1)		; leemos dato
		jind	r6			; y retornamos


;----------------------------------------------
; Imprime R0 en hexadecimal
prthex:
		; prólogo
		subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r3
		;
		or		r2,r0,r0
		ldi		r3,4
prth1:	add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2
		add		r0,r2,r2	; rol sin carry
		adc		r2,r2,r2	
		andi	r0,r2,15	; Los bits interesantes están en LSB
		cmpi	r0,10
		jnc		prth2
		addi	r0,r0,7
prth2:	ldi		r1,'0'
		add		r0,r0,r1
		adpc 	r6,1
		jr		putch
		subi	r3,r3,1
		jnz		prth1
		; epílogo
		ld		r3,(r7)
		addi	r7,r7,1
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;---------------------------------------------------
; Imprime R0 como 2 caracteres ASCII (little-endian)
; si es que son imprimibles. En caso de no serlo
; imprime '.'
putdchle:
		subi	r7,r7,1
		st		(r7),r6

		shr		r2,r0		; Ahorramos un desplazamiento
		ldi		r1,0xff
		and		r0,r0,r1
		ldi		r1,32
		cmp		r0,r1
		jc		pdch1
		ldi		r0,'.'
pdch1:	ldi		r1,127
		cmp		r0,r1
		jnc		pdch2
		ldi		r0,'.'
pdch2:	adpc 	r6,1
		jr		putch
		shr		r0,r2
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		ldi		r1,32
		cmp		r0,r1
		jc		pdch3
		ldi		r0,'.'
pdch3:	ldi		r1,127
		cmp		r0,r1
		jnc		pdch4
		ldi		r0,'.'
pdch4:	adpc 	r6,1
		jr		putch

		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6

;----------------------------------------------
; Hexadecimal dump
; R0: dir base
; R1: puntero a datos
; R2: Nº de filas de datos (8 datos por fila)

hexdump:
		; prólogo
		subi	r7,r7,1
		st		(r7),r6
		subi	r7,r7,1
		st		(r7),r3
		subi	r7,r7,1
		st		(r7),r4
		subi	r7,r7,1
		st		(r7),r5
		;
		or		r3,r0,r0	; r3: dir base
		or		r4,r1,r1	; r4: puntero a datos
		or		r5,r2,r2	; r5: contador de filas
hd1:	or		r0,r3,r3	; Imprimimos direccion
		adpc 	r6,1
		jr		prthex
		ldi		r0,' '		; y dos espacios
		adpc 	r6,1
		jr		putch
		ldi		r0,' '
		adpc 	r6,1
		jr		putch

		subi	r7,r7,1		; Copia temporal de puntero
		st		(r7),r4
hd2:	ld		r0,(r4)		; Imprimimos los datos
		addi	r4,r4,1
		adpc 	r6,1
		jr		prthex
		ldi		r0,' '
		adpc 	r6,1
		jr		putch
		andi	r0,r4,7
		jnz		hd2

		ldi		r0,' '		; y un espacio
		adpc 	r6,1
		jr		putch

		ld		r4,(r7)		; recuperamos el puntero
		addi	r7,r7,1		; para repetir los datos en ASCII little endian
hd3:	ld		r0,(r4)
		adpc 	r6,1
		jr		putdchle
		addi	r4,r4,1	
		andi	r0,r4,7
		jnz		hd3

		addi	r3,r3,8		; nueva línea
		ldi		r0,'\n'
		adpc 	r6,1
		jr		putch
		subi	r5,r5,1
		jnz		hd1

		; epílogo
		ld		r3,(r7)
		addi	r7,r7,1
		ld		r4,(r7)
		addi	r7,r7,1
		ld		r5,(r7)
		addi	r7,r7,1
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6



txtle:	asczle	"\n\nPrueba de Flag V\n"

outreg:	word 0x10
		org	(.+7)&(~7)	; alineamos a múltiplo de 8
buffer:



