;----------- REGISTROS de E/S -----------

IRQEN=		0x20
TIMER=		0x21
UARTDAT=	0x22
SPIDAT=		0x23
PFLAGS=		0x24
SPICTRL= 	0x25
GPIN=		0x25
GPDIR=		0x26
GPOUT=		0x27

;--------- Habilitación de interrupciones --------
EINTEX=		1		; Habilita IRQ externa
ETIMER=		2		; Habilita IRQ Timer
EUARTTHRE=	4		; Habilita IRQ UART TX
EUARTDV=	8		; Habilita IRQ UART RX

;--------- bits de PFLAGS -----------
UARTDV=		1		; Dato válido en UART RX
UARTTHRE=	2		; Listo para transmitir en UART TX
UARTTEND=	4		; Transmisión completa en UART TX
SPIBUSY=	8		; SPI ocupado transfiriendo datos
UARTFE=		0x10	; Error de trama (bit de STOP en 0) en UART RX
UARTOVE=	0x20	; Error de overrun en UART RX
TIMERIRQ=	0x8000	; Flag de IRQ del temporizador

;------------- Notas E/S --------------------
; - La escritura en TIMER carga el valor máximo de la cuenta y pone el contador en 0
; - La lectura de TIMER devuelve el valor actual de la cuenta y borra la petición de interrupción
; - El bit 8 de SPICTRL selecciona el modo de 16 bits si 1 o 8 bits si 0.
; - Los bits 7:0 de SPICTRL seleccionan la velocidad de SCK como f_sck=f_clk/(2*(SPICTRL[7:0]+1))

		org 0			; RESET
;		jr	inicio
		ldh		>inicio
		ldi		r0,<inicio
		jind	r0

		org 4			; IRQ0: externa
		JR		IRQ0
		org 8			; IRQ1: TIMER
		jr		IRQ1
		org	0xc			; IRQ2:	UART TX
		jr		IRQ2
		org	0x10		; IRQ3: UART RX
		jr		IRQ3

		org	64
IRQ1:	subi	r7,r7,1
		st		(r7),r0
		subi	r7,r7,1
		st		(r7),r1

		ldi		r1,TIMER	; La lectura borra la IRQ del timer
		ld		r0,(r1)

		ldh		>outreg
		ldi		r1,<outreg
		ld		r0,(r1)
		addi	r0,r0,1
		st		(r1),r0
		ldi		r1,<GPOUT
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		st		(r1),r0

		ld		r1,(r7)
		addi	r7,r7,1
		ld		r0,(r7)
		addi	r7,r7,1
IRQ0:
IRQ2:
IRQ3:
		reti

inicio:
		ldi		r1,0
		ldi		r2,0
		ldi		r3,0
		ldi		r4,0
		ldi		r5,0
		ldi		r6,0

		ldh 	>4096		; Puntero de Pila al final de RAM
		ldi 	r7,<4096

		; Prueba SPI 
		ldi		r1,SPICTRL
		ldi		r0,0	; SSB todas en alto
		st		(r1),r0
		ldh		0x81	; Modo 16 bits, SS1
		ldi		r0,0	; Velocidad máxima CLK/(2*(val+1)) = CLK/2
		st		(r1),r0

		ldi		r1,SPIDAT
		st		(r1),r0

		ldi		r1,PFLAGS
bspi1:	ld		r0,(r1)
		tsti	r0,SPIBUSY
		jnz		bspi1

		ldi		r1,SPIDAT
		ld		r0,(r1)		; lectura de RX 

		ldi		r1,SPICTRL
		ldh		2			; Modo 8 bits, SS2
		ldi		r0,4		; velocidad CLK/(2*(val+1)) = CLK/10
		st		(r1),r0

		ldi		r1,SPIDAT
		st		(r1),r0

		ldi		r1,PFLAGS
bspi2:	ld		r0,(r1)
		tsti	r0,SPIBUSY
		jnz		bspi2

		ldi		r1,SPIDAT
		ld		r0,(r1)		; lectura de RX 

		ldi		r1,SPICTRL
		ldh		3			; Modo 8 bits, SS3
		ldi		r0,0		; velocidad CLK/(2*(val+1)) = CLK/2
		st		(r1),r0
		
		; GPIO como salida
		ldi		r1,GPDIR
		ldi		r0,0x1F
		st		(r1),r0 

		; Prueba de interrupciones 

		ldi		r1,TIMER	; Interrupción cada 100 ciclos
		ldh		>9999
		ldi		r0,<9999
		st		(r1),r0
		ld		r0,(r1)		; Borramos IRQ
		ldi		r1,IRQEN	; Habilita IRQ Timer
		ldi		r0,ETIMER
		st		(r1),r0


		ldh	>txtle		; Imprime cadena de caracteres TXT
		ldi r0,<txtle	; empaquetada tipo Little-Endian
		adpc r6,1
		jr	putsle

buc:	adpc r6,1
		jr	getch
		ldi	r1,'a'
		cmp r0,r1		; Carry en 0 si llevada al restar (r1>=r0 o r0 mayúscula)
		jnc	noch
		ldi	r1,0x20		; Cambiar bit 5
		xor	r0,r0,r1
noch:	;ldi	r1,GPOUT
		;st	(r1),r0
		adpc r6,1
		jr	putch
		jr	buc

;----------------------------------------------------
; Envía cadena de caracteres empaquetados al terminal
; parámetros:
;	R0: puntero a la cadena de caracteres
;	R6: dirección de retorno
; retorna:
; 	R0-R2: modificados

putsle:	subi	r7,r7,1
		st		(r7),r6

		or		r2,r0,r0
putsle1:
		ld		r0,(r2)
		ldi		r1,0xff
		and		r0,r0,r1
		jz		putsle3
		adpc	r6,1
		jr		putch

		ld		r0,(r2)
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		shr		r0,r0
		jz		putsle3
		adpc	r6,1
		jr		putch
		addi	r2,r2,1
		jr		putsle1

putsle3:
		ld		r6,(r7)
		addi	r7,r7,1
		jind	r6
;----------------------------------------------
; Envía caracter al terminal
; parámetros:
;	R0: dato a enviar
;	R6: dirección de retorno
; retorna:
; 	R1: modificado

putch:	ldi		r1,PFLAGS	; Esperamos flag TBE en 1
		ld		r1,(r1)		
		tsti	r1,UARTTHRE
		jz		putch
		ldi		r1,<UARTDAT		
		st		(r1),r0		; envía dato
		jind	r6			; y retornamos

;----------------------------------------------
; Recibe caracter del terminal
; parámetros:
;	R6: dirección de retorno
; retorna:
;	R0: dato recibido
; 	R1: modificado

getch:	ldi		r1,PFLAGS
getch1:	ld		r0,(r1)		; esperamos flag DV en 1
		tsti	r0,UARTDV
		jz		getch1
		ldi		r1,UARTDAT	; ahora apuntamos a registro RX
		ld		r0,(r1)		; leemos dato
		jind	r6			; y retornamos

txtle:	asczle	"\n\n¡¡¡Hola Mundo!!!\n"

outreg:	word 0x10


