// Temporizador con salida PWM
// pero para PWM sólo se usan 10 bits

module TIMERPWM(
  input clk,      // Reloj principal
  input [15:0]d,  // Datos desde la CPU
  input wrtim,    // strobe de escritura en timer
  input wrpwm,	  // strobe de escritura en PWM (borra IRQ)
  input rdtim,	  // strobe de lectura (borra IRQ)
  output [15:0]q, // salida hacia CPU
  output irq,	  // Salida de petición de interrupción
  output pwmp,	  // Salida PWM	positiva
  output pwmn	  // Salida PWM negativa
);

reg [15:0]timerval=0;	// Contador
reg [15:0]timermax=0;	// Valor máximo
reg [10:0] hold=0;		// registro temporal de valor
reg [10:0] curr=0;		// registro de valor del ciclo PWM actual
reg pwm=0;				// Salida PWM
reg irq=0;				// Salida de petición de interrupción
wire max;				// Final de cuenta
reg maxr;				// Final de cuenta retrasado 1 ciclo
wire mat;				// matching entre cont y curr
wire nciclo;			// comienzo de ciclo

wire [9:0] abs;			// valor absoluto de la muestra

assign max = (timerval==timermax);
assign mat = (timerval[9:0]==curr[9:0]);

assign abs = hold[10] ? ((~hold[9:0])+1) : hold[9:0];

always @ (posedge clk)
begin
	maxr <= max;
	timermax <= wrtim ? d : timermax;
	timerval <= (wrtim|max) ? 0 : timerval+1;
	hold <= wrpwm ? d[10:0] : hold;
	curr <= max ? {hold[10],abs} : curr;
	irq  <= max ? 1 : ((wrpwm|rdtim) ? 0 : irq);
	pwm  <= mat ? 0 : (maxr ? 1 : pwm);
end

assign q = timerval;

assign	pwmp = pwm & (~curr[10]);
assign	pwmn = pwm & ( curr[10]);


endmodule

